/*
 * audio_transitions.cc -- audio transitions
 * Copyright (C) 2002 Charles Yates <charles.yates@pandora.be>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "audio_transitions.h"
#include "kino_av_pipe.h"
#include "frame.h"

extern "C" {
#include "magick_interface.h"
#include "support.h"
}

/** Audio transition switch - determines the point at which the b frames audio is used during
	an frame transition.
*/

class AudioSwitch : public GDKAudioTransition
{
	private:
		GtkWidget *window;
		double switch_point;

	public:
		AudioSwitch( ) 
		{
			window = create_window_switch( );
		}

		virtual ~AudioSwitch( ) 
		{
			gtk_widget_destroy( window );
		}

		char *GetDescription( ) const 
		{ 
			return "Switch";
		}

		void GetFrame( int16_t **aframe, int16_t **bframe, int frequency, int channels, int samples, double position, double frame_delta )
		{
			if ( switch_point < position )
				for ( int i = 0; i < channels; i ++ )
					memcpy( aframe[ i ], bframe[ i ], samples * sizeof( int16_t ) );
		}

		void AttachWidgets( GtkBin *bin ) 
		{
			gtk_widget_reparent( ( GTK_BIN( window ) )->child, GTK_WIDGET( bin ) );
		}

		void DetachWidgets( GtkBin *bin ) 
		{ 
			gtk_widget_reparent( ( GTK_BIN( bin ) )->child, GTK_WIDGET( window ) );
		}

		void InterpretWidgets( GtkBin *bin ) 
		{ 
			GtkRange *range = GTK_RANGE( lookup_widget( window, "hscale_switch" ) );
			GtkAdjustment *adjust = gtk_range_get_adjustment( range );
			switch_point = (double)adjust->value / 100; 
		}
};

/** Helper class for transitions with wav inputs associated to them. This class encapsulates
	the KinoAudioInput and resampling requirements.
*/

static void on_entry_file_changed( GtkEditable *editable, gpointer user_data );

class WavSelect
{
	private:
		char wav_file[ PATH_MAX + NAME_MAX ];
		KinoAudioInput *wav;
		GtkLabel *label_info;
		GtkEntry *file_entry;
		FastAudioResample resampler;
		int frame_size;
		int fps;
		int16_t temp[ 4 * DV_AUDIO_MAX_SAMPLES ];
		bool wav_selected;

	public:
		WavSelect() : wav(NULL), label_info(NULL), file_entry(NULL), wav_selected(false)
		{
			strcpy( wav_file, "" );
		}

		~WavSelect()
		{
			delete wav;
		}

		void SetWavInfoLabel( GtkLabel *label )
		{
			label_info = label;
		}

		void SetWavFileEntry( GtkEntry *entry )
		{
			file_entry = entry;
			gtk_signal_connect( GTK_OBJECT( entry ), "changed", GTK_SIGNAL_FUNC( on_entry_file_changed ), this );
		}

		void WavFileSelected( )
		{
			char text[ 10240 ] = "";

			if ( strcmp( wav_file, gtk_entry_get_text( file_entry ) ) ) 
			{
				if ( this->wav != NULL )
					delete wav;

				strcpy( wav_file, gtk_entry_get_text( file_entry ) );
				this->wav = KinoAudioInputFactory::CreateAudioInput( wav_file );

				if ( wav != NULL )
				{
					wav_selected = true;
					sprintf( text, "Selected file has a frequency of %d Hz\nand lasts %.02f seconds", 
							wav->GetFrequency(), (double)wav->GetNumberOfSamples() / (double)wav->GetFrequency() );
				}
				else
				{
					wav_selected = false;
					strcpy( text, "Invalid file selected" );
				}
			}
			else if ( !strcmp( wav_file, "" ) )
			{
				wav_selected = false;
				strcpy( text, "No file selected" );
			}

			if ( strcmp( text, "" ) )
				gtk_label_set_text( label_info, text );
		}

		void WavStart( int frequency, int samples, int offset )
		{
			if ( wav_selected )
			{
				resampler.SetOutputFrequency( frequency );
				fps = frequency / samples;
				frame_size = wav->GetChannels() * wav->GetBytesPerSample() * wav->GetFrequency() / ( fps );
				wav->Seek( offset * frame_size );
			}
		}

		void WavRead( int16_t **a, int f, int c )
		{
			if ( wav_selected )
			{
				if ( !wav->Get( temp, frame_size ) )
				{
					printf( ">>> Underrun? unable to read %d bytes\n", frame_size );
					memset( temp, 0, sizeof( temp ) );
				}

				resampler.Resample( temp, wav->GetFrequency(), c, frame_size );

				for ( int s = 0; s < f / fps; s ++ )
					for ( int i = 0; i < c; i ++ )
						a[ i ][ s ] = resampler.output[ s * c + i ];
			}
		}

		bool IsSelected()
		{
			return wav_selected;
		}
};

/** GTK callback on file selection change.
*/

static void on_entry_file_changed( GtkEditable *editable, gpointer user_data )
{
	( ( WavSelect * ) user_data )->WavFileSelected();
}

/** Dub the video with the contents of a wav. This transition is about as un-transitiony as you get and would
	be better suited to the audio filters. Still it serves its purpose for now...

	NB: To avoid crashes during a preview which starts without audio and then has it switched on
	midflow, the initiated flag is used to determine that audio is initialised correctly at the
	start of the effect... This has the unfortunate effect of making audio require a restart in
	the preview before it is actually played correctly...
*/

class AudioDub : public GDKAudioTransition, WavSelect
{
	private:
		GtkWidget *window;
		int offset;
		bool initiated;

	public:
		AudioDub( ) : initiated( false )
		{
			window = create_window_dub( );
			SetWavInfoLabel( GTK_LABEL( lookup_widget( window, "label_dub_info" ) ) );
			SetWavFileEntry( GTK_ENTRY( lookup_widget( window, "entry_dub" ) ) );
			WavFileSelected();
		}

		virtual ~AudioDub( ) 
		{
		}

		char *GetDescription( ) const 
		{ 
			return "Dub";
		}

		void GetFrame( int16_t **aframe, int16_t **bframe, int frequency, int channels, int samples, double position, double frame_delta )
		{
			if ( position == 0 )
			{
				initiated = true;
				WavStart( frequency, samples, offset );
			}

			if ( initiated )
				WavRead( aframe, frequency, channels );
		}

		void AttachWidgets( GtkBin *bin ) 
		{
			gtk_widget_reparent( ( GTK_BIN( window ) )->child, GTK_WIDGET( bin ) );
		}

		void DetachWidgets( GtkBin *bin ) 
		{ 
			gtk_widget_reparent( ( GTK_BIN( bin ) )->child, GTK_WIDGET( window ) );
		}

		void InterpretWidgets( GtkBin *bin ) 
		{
			if ( !IsSelected() )
				throw "No audio input selected for dubbing - aborting.";

			GtkEntry *offsetSpin = GTK_ENTRY( lookup_widget( window, "spinbutton_dub_offset" ) );
			offset = atoi( gtk_entry_get_text( offsetSpin ) );

			initiated = false;
		}

};

/** Mix the audio with a wav file. This is a proper 'audio transition' in that it takes care of the switch
	from a to b frames.

	NB: To avoid crashes during a preview which starts without audio and then has it switched on
	midflow, the initiated flag is used to determine that audio is initialised correctly at the
	start of the effect... This has the unfortunate effect of making audio require a restart in
	the preview before it is actually played correctly...
*/

class AudioMix : public GDKAudioTransition, WavSelect
{
	private:
		GtkWidget *window;
		bool initiated;
		int offset;
		double mix_ratio;
		double switch_point;

	public:
		AudioMix( ) : initiated( false )
		{
			window = create_window_mix( );
			SetWavInfoLabel( GTK_LABEL( lookup_widget( window, "label_mix_info" ) ) );
			SetWavFileEntry( GTK_ENTRY( lookup_widget( window, "entry_mix" ) ) );
			WavFileSelected();
		}

		virtual ~AudioMix( ) 
		{
		}

		char *GetDescription( ) const 
		{ 
			return "Mix";
		}

		void Mix( int16_t **io, int16_t **with, int channels, int samples )
		{
			for ( int s = 0; s < samples; s ++ )
				for ( int c = 0; c < channels; c ++ )
					io[ c ][ s ] = (int16_t)((double)io[c][s] * mix_ratio + (double)with[c][s] * ( 1 - mix_ratio ) );
		}

		void GetFrame( int16_t **aframe, int16_t **bframe, int frequency, int channels, int samples, double position, double frame_delta )
		{ 
			if ( position == 0 )
			{
				initiated = true;
				WavStart( frequency, samples, offset );
			}

			if ( initiated && position >= switch_point )
			{
				WavRead( aframe, frequency, channels );
				Mix( bframe, aframe, channels, samples );
				for ( int index = 0; index < channels; index ++ )
					memcpy( aframe[ index ], bframe[ index ], samples * sizeof( int16_t ) );
			}
			else if ( initiated )
			{
				WavRead( bframe, frequency, channels );
				Mix( aframe, bframe, channels, samples );
			}
		}

		void AttachWidgets( GtkBin *bin ) 
		{
			gtk_widget_reparent( ( GTK_BIN( window ) )->child, GTK_WIDGET( bin ) );
		}

		void DetachWidgets( GtkBin *bin ) 
		{ 
			gtk_widget_reparent( ( GTK_BIN( bin ) )->child, GTK_WIDGET( window ) );
		}

		void InterpretWidgets( GtkBin *bin ) 
		{ 
			if ( !IsSelected() )
				throw "No audio input selected for mixing - aborting.";

			GtkEntry *offsetSpin = GTK_ENTRY( lookup_widget( window, "spinbutton_mix_offset" ) );
			offset = atoi( gtk_entry_get_text( offsetSpin ) );

			GtkRange *range = GTK_RANGE( lookup_widget( window, "hscale_mix_ratio" ) );
			GtkAdjustment *adjust = gtk_range_get_adjustment( range );
			mix_ratio = (double)adjust->value / 100; 

			range = GTK_RANGE( lookup_widget( window, "hscale_mix_switch" ) );
			adjust = gtk_range_get_adjustment( range );
			switch_point = (double)adjust->value / 100;

			initiated = false;
		}
};

/** Callback for selection change.
*/

static void
on_optionmenu_selected (GtkMenuItem *menu_item, gpointer user_data) 
{
	( (GDKAudioTransitionRepository *) user_data )->SelectionChange();
}

/** Constructor for the audio transition repository. 

  	Registers an instance of each known transition for later GUI exposure via the Initialise method.
*/

GDKAudioTransitionRepository::GDKAudioTransitionRepository() : selected_transition( NULL ), menu( NULL ), container( NULL )
{
	printf( ">> audio transition repository created\n" );
	// Register an instance of each object (adapting raw transitions to GDK transitions where necessary)
	Register( new AudioSwitch() );
	Register( new AudioDub() );
	Register( new AudioMix() );
}

/** Destructor for the image repository - clears all the registered transitions
*/

GDKAudioTransitionRepository::~GDKAudioTransitionRepository()
{
	printf( ">> audio transition repository destroyed\n" );
	// Remove the transitions in the repository
	for ( unsigned int index = 0; index < transitions.size(); index ++ )
		delete transitions[ index ];
}

/** Register an audio transition.
*/

void GDKAudioTransitionRepository::Register( GDKAudioTransition *transition )
{
	printf( ">>> Audio Transition: %s\n", transition->GetDescription() );
	transitions.push_back( transition );
}

/** Initialise the option menu with the current list of registered transitions.
*/

void GDKAudioTransitionRepository::Initialise( GtkOptionMenu *menu, GtkBin *container )
{
	// Store these for future reference
	this->menu = menu;
	this->container = container;

	// Add the transitions to the menu
	GtkMenu *menu_new = GTK_MENU( gtk_menu_new( ) );
	for ( unsigned int index = 0; index < transitions.size(); index ++ )
	{
		GtkWidget *item = gtk_menu_item_new_with_label( transitions[ index ]->GetDescription( ) );
		gtk_widget_show( item );
		gtk_menu_append( menu_new, item );
        gtk_signal_connect( GTK_OBJECT( item ), "activate", GTK_SIGNAL_FUNC( on_optionmenu_selected ), this );
	}
	gtk_menu_set_active( menu_new, 0 );
	gtk_option_menu_set_menu( menu, GTK_WIDGET( menu_new ) );

	// Register the selected items widgets
	SelectionChange();
}

/** Get the currently selected audio transition.
*/

GDKAudioTransition *GDKAudioTransitionRepository::Get( ) const
{
	GtkMenu *transitionMenu = GTK_MENU( gtk_option_menu_get_menu( menu ) );
	GtkWidget *active_item = gtk_menu_get_active( transitionMenu );
	return transitions[ g_list_index( GTK_MENU_SHELL( transitionMenu )->children, active_item ) ];
}

/** Handle attach/detach widgets on last selected/selected items.
*/

void GDKAudioTransitionRepository::SelectionChange( )
{
	// Detach the selected transitions widgets
	if ( selected_transition != NULL )
		selected_transition->DetachWidgets( container );

	// Get the new selection
	selected_transition = Get();

	// Attach the new transitions widgets
	if ( selected_transition != NULL )
		selected_transition->AttachWidgets( container );
}

