/*
 * avi.cc library for AVI file format i/o
 * Copyright (C) 2000 - 2002 Arne Schirmacher <arne@schirmacher.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Tag: $Name:  $
 *
 * Change log:
 * 
 * $Log: avi.cc,v $
 * Revision 1.15  2002/06/10 10:39:51  ddennedy
 * minor fixes for large files
 *
 * Revision 1.14  2002/05/17 08:04:24  ddennedy
 * revert const-ness of Frame references in Frame, FileHandler, and AVI classes
 *
 * Revision 1.13  2002/05/15 04:39:35  ddennedy
 * bugfixes to dv2 AVI write, audio export, Xv init
 *
 * Revision 1.12  2002/04/29 05:09:21  ddennedy
 * raw dv file support, Frame::ExtractAudio uses libdv, audioScrub prefs
 *
 * Revision 1.11  2002/04/15 19:12:32  schirmacher
 * removed debugging code causing performance losses and crashes with dv2 files
 *
 * Revision 1.10  2002/04/09 06:53:42  ddennedy
 * cleanup, new libdv 0.9.5, large AVI, dnd storyboard
 *
 * Revision 1.8  2002/03/25 21:34:25  arne
 * Support for large (64 bit) files mostly completed
 *
 * Revision 1.7  2002/03/10 21:28:29  arne
 * release 1.1b1, 64 bit support for type 1 avis
 *
 * Revision 1.6  2002/03/10 13:29:41  arne
 * more changes for 64 bit access
 *
 * Revision 1.5  2002/03/09 17:59:28  arne
 * moved index routines to AVIFile
 *
 * Revision 1.4  2002/03/09 10:26:26  arne
 * improved constructors and assignment operator
 *
 * Revision 1.3  2002/03/09 08:55:57  arne
 * moved a few variables to AVIFile
 *
 * Revision 1.2  2002/03/04 19:22:43  arne
 * updated to latest Kino avi code
 *
 * Revision 1.1.1.1  2002/03/03 19:08:08  arne
 * import of version 1.01
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// C includes

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>

// C++ includes

#include <string>
#include <iostream>
#include <strstream>
#include <iomanip>

using std::cout;
using std::hex;
using std::dec;
using std::setw;
using std::setfill;
using std::endl;

// local includes

#include "error.h"
#include "riff.h"
#include "avi.h"

#define PADDING_SIZE (512)
#define IX00_INDEX_SIZE (4028)

#define AVIF_HASINDEX 0x00000010
#define AVIF_MUSTUSEINDEX 0x00000020
#define AVIF_TRUSTCKTYPE 0x00000800
#define AVIF_ISINTERLEAVED 0x00000100
#define AVIF_WASCAPTUREFILE 0x00010000
#define AVIF_COPYRIGHTED 0x00020000


static char g_zeroes[PADDING_SIZE];

/** The constructor
 
    \todo mainHdr not initialized
    \todo add checking for NULL pointers
 
*/

AVIFile::AVIFile() : RIFFFile(), 
        idx1( NULL), file_list( -1), riff_list( -1),
        hdrl_list( -1), avih_chunk( -1), movi_list( -1), junk_chunk( -1), idx1_chunk( -1),
        index_type( -1), current_ix00( -1)
{
    // cerr << "0x" << hex << (long)this << dec << " AVIFile::AVIFile() : RIFFFile(), ..." << endl;

    for (int i = 0; i < 2; ++i) {
	indx[i] = new AVISuperIndex;
        memset(indx[i], 0, sizeof(AVISuperIndex));
        ix[i] = new AVIStdIndex;
        memset(ix[i], 0, sizeof(AVIStdIndex));
        indx_chunk[i] = -1;
        ix_chunk[i] = -1;
        strl_list[i] = -1;
        strh_chunk[i] = -1;
        strf_chunk[i] = -1;
    }
    idx1 = new AVISimpleIndex;
    memset(idx1, 0, sizeof(AVISimpleIndex));
}


/** The copy constructor
 
    \todo add checking for NULL pointers
 
*/

AVIFile::AVIFile(const AVIFile& avi) : RIFFFile(avi)
{
    // cerr << "0x" << hex << (long)this << dec << " 0x" << hex << (long)&avi << dec << " AVIFile::AVIFile(const AVIFile& avi) : RIFFFile(avi)" << endl;

    mainHdr = avi.mainHdr;
    idx1 = new AVISimpleIndex;
    *idx1 = *avi.idx1;
    file_list = avi.file_list;
    riff_list = avi.riff_list;
    hdrl_list = avi.hdrl_list;
    avih_chunk = avi.avih_chunk;
    movi_list = avi.movi_list;
    junk_chunk = avi.junk_chunk;
    idx1_chunk = avi.idx1_chunk;

    for (int i = 0; i < 2; ++i) {
        indx[i] = new AVISuperIndex;
        *indx[i] = *avi.indx[i];
        ix[i] = new AVIStdIndex;
        *ix[i] = *avi.ix[i];
        indx_chunk[i] = avi.indx_chunk[i];
        ix_chunk[i] = avi.ix_chunk[i];
        strl_list[i] = avi.strl_list[i];
        strh_chunk[i] = avi.strh_chunk[i];
        strf_chunk[i] = avi.strf_chunk[i];
    }

    index_type = avi.index_type;
    current_ix00 = avi.current_ix00;
}


/** The assignment operator
 
*/

AVIFile& AVIFile::operator=(const AVIFile& avi)
{
    // cerr << "0x" << hex << (long)this << dec << " 0x" << hex << (long)&avi << dec << " AVIFile& AVIFile::operator=(const AVIFile& avi)" << endl;

    if (this != &avi) {
        RIFFFile::operator=(avi);
        mainHdr = avi.mainHdr;
        *idx1 = *avi.idx1;
        file_list = avi.file_list;
        riff_list = avi.riff_list;
        hdrl_list = avi.hdrl_list;
        avih_chunk = avi.avih_chunk;
        movi_list = avi.movi_list;
        junk_chunk = avi.junk_chunk;
        idx1_chunk = avi.idx1_chunk;

        for (int i = 0; i < 2; ++i) {
            *indx[i] = *avi.indx[i];
            *ix[i] = *avi.ix[i];
            indx_chunk[i] = avi.indx_chunk[i];
            ix_chunk[i] = avi.ix_chunk[i];
            strl_list[i] = avi.strl_list[i];
            strh_chunk[i] = avi.strh_chunk[i];
            strf_chunk[i] = avi.strf_chunk[i];
        } 

        index_type = avi.index_type;
        current_ix00 = avi.current_ix00;
    }
    return *this;
}


/** The destructor
 
*/

AVIFile::~AVIFile()
{
    // cerr << "0x" << hex << (long)this << dec << " AVIFile::~AVIFile()" << endl;

    for (int i = 0; i < 2; ++i) {
        delete ix[i];
        delete indx[i];
    }
    delete idx1;
}

/** Initialize the AVI structure to its initial state, either for PAL or NTSC format
 
    Initialize the AVIFile attributes: mainHdr, indx, ix00, idx1
 
    \todo consolidate AVIFile::Init, AVI1File::Init, AVI2File::Init. They are somewhat redundant.
    \param format pass AVI_PAL or AVI_NTSC
    \param sampleFrequency the sample frequency of the audio content
    \param indexType pass AVI_SMALL_INDEX or AVI_LARGE_INDEX
 
*/

void AVIFile::Init(int format, int sampleFrequency, int indexType)
{
    int i, j;

    assert((format == AVI_PAL) || (format == AVI_NTSC));

    index_type = indexType;

    switch (format) {
    case AVI_PAL:
        mainHdr.dwMicroSecPerFrame = 40000;
        mainHdr.dwSuggestedBufferSize = 144008;
        break;

    case AVI_NTSC:
        mainHdr.dwMicroSecPerFrame = 33366;
        mainHdr.dwSuggestedBufferSize = 120008;
        break;

    default: /* no default allowed */
        assert(0);
        break;
    }

    /* Initialize the 'avih' chunk */

    mainHdr.dwMaxBytesPerSec = 3600000;
    mainHdr.dwPaddingGranularity = PADDING_SIZE;
    mainHdr.dwFlags = AVIF_HASINDEX | AVIF_TRUSTCKTYPE;
    mainHdr.dwTotalFrames = 0;
    mainHdr.dwInitialFrames = 0;
    mainHdr.dwStreams = 1;
    mainHdr.dwWidth = 0;
    mainHdr.dwHeight = 0;
    mainHdr.dwReserved[0] = 0;
    mainHdr.dwReserved[1] = 0;
    mainHdr.dwReserved[2] = 0;
    mainHdr.dwReserved[3] = 0;

    /* Initialize the 'idx1' chunk */

    for (int i = 0; i < 8000; ++i) {
        idx1->aIndex[i].dwChunkId = 0;
        idx1->aIndex[i].dwFlags = 0;
        idx1->aIndex[i].dwOffset = 0;
        idx1->aIndex[i].dwSize = 0;
    }
    idx1->nEntriesInUse = 0;

    /* Initialize the 'indx' chunk */

    for (i = 0; i < 2; ++i) {
        indx[i]->wLongsPerEntry = 4;
        indx[i]->bIndexSubType = 0;
        indx[i]->bIndexType = AVI_INDEX_OF_INDEXES;
        indx[i]->nEntriesInUse = 0;
        indx[i]->dwReserved[0] = 0;
        indx[i]->dwReserved[1] = 0;
        indx[i]->dwReserved[2] = 0;
        for (j = 0; j < 2014; ++j) {
            indx[i]->aIndex[j].qwOffset = 0;
            indx[i]->aIndex[j].dwSize = 0;
            indx[i]->aIndex[j].dwDuration = 0;
        }
    }

    /* The ix00 and ix01 chunk will be added dynamically in avi_write_frame
              as needed */
}


/** Find position and size of a given frame in the file
 
    Depending on which index is available, search one of them to
    find position and frame size
 
    \todo the size parameter is redundant. All frames have the same size, which is also in the mainHdr.
    \todo all index related operations should be isolated 
    \todo offset should be 64 bit
    \bug this won't work with very long AVI files, because these have several indexes
    \param offset the file offset to the start of the frame
    \param size the size of the frame
    \param the number of the frame we wish to find
    \return 0 if the frame could be found, -1 otherwise
*/

int AVIFile::GetDVFrameInfo(off_t &offset, int &size, int frameNum)
{
    switch (index_type) {
    case AVI_LARGE_INDEX:

        /* find relevant index in indx0 */

        int i;

        for (i = 0; frameNum >= indx[0]->aIndex[i].dwDuration; frameNum -= indx[0]->aIndex[i].dwDuration, ++i)
            ;

        if (i != current_ix00) {
            fail_if(lseek(fd, indx[0]->aIndex[i].qwOffset + RIFF_HEADERSIZE, SEEK_SET) == (off_t)-1);
            fail_neg(read(fd, ix[0], indx[0]->aIndex[i].dwSize - RIFF_HEADERSIZE));
            current_ix00 = i;
        }

        if (frameNum < ix[0]->nEntriesInUse) {
            offset = ix[0]->qwBaseOffset + ix[0]->aIndex[frameNum].dwOffset;
            size = ix[0]->aIndex[frameNum].dwSize;
            return 0;
        } else
            return -1;
        break;

    case AVI_SMALL_INDEX:
        int index = -1;
        int frameNumIndex = 0;
        FOURCC chunkId = make_fourcc("00dc");
        for (int i = 0; i < idx1->nEntriesInUse; ++i) {
            if (idx1->aIndex[i].dwChunkId == chunkId) {
                if (frameNumIndex == frameNum) {
                    index = i;
                    break;
                }
                ++frameNumIndex;
            }
        }
        if (index != -1) {
            // compatibility check for broken dvgrab dv2 format
            if (idx1->aIndex[0].dwOffset > GetDirectoryEntry(movi_list).offset) {
                offset = idx1->aIndex[index].dwOffset + RIFF_HEADERSIZE;
            } else {
                // new, correct dv2 format
                offset = idx1->aIndex[index].dwOffset + RIFF_HEADERSIZE + GetDirectoryEntry(movi_list).offset;
            }
            size = idx1->aIndex[index].dwSize;
            return 0;
        } else
            return -1;
        break;
    }
    return -1;
}


/** Read in a frame
 
    \todo use 64 bit seek
    \todo we actually don't need the frame here, we could use just a void pointer
    \param frame a reference to the frame object that will receive the frame data
    \param frameNum the frame number to read
    \return 0 if the frame could be read, -1 otherwise
*/

int AVIFile::GetDVFrame(Frame &frame, int frameNum)
{
    off_t	offset;
    int	size;

    if (GetDVFrameInfo(offset, size, frameNum) != 0)
        return -1;
    fail_if(lseek(fd, offset, SEEK_SET) == (off_t)-1);
    fail_neg(read(fd, frame.data, size));

    return 0;
}


int AVIFile::GetTotalFrames() const
{
    return mainHdr.dwTotalFrames;
}


/** prints out a directory entry in text form
 
    Every subclass of RIFFFile is supposed to override this function
    and to implement it for the entry types it knows about. For all
    other entry types it should call its parent::PrintDirectoryData.
 
    \todo use 64 bit routines
    \param entry the entry to print
*/

void AVIFile::PrintDirectoryEntryData(const RIFFDirEntry &entry) const
{
    if (entry.type == make_fourcc("avih")) {

        int             i;
        MainAVIHeader   main_avi_header;

        fail_if(lseek(fd, entry.offset, SEEK_SET) == (off_t)-1);
        fail_neg(read(fd, &main_avi_header, sizeof(MainAVIHeader)));

        cout << "    dwMicroSecPerFrame:    " << (int)main_avi_header.dwMicroSecPerFrame << endl
        << "    dwMaxBytesPerSec:      " << (int)main_avi_header.dwMaxBytesPerSec << endl
        << "    dwPaddingGranularity:  " << (int)main_avi_header.dwPaddingGranularity << endl
        << "    dwFlags:               " << (int)main_avi_header.dwFlags << endl
        << "    dwTotalFrames:         " << (int)main_avi_header.dwTotalFrames << endl
        << "    dwInitialFrames:       " << (int)main_avi_header.dwInitialFrames << endl
        << "    dwStreams:             " << (int)main_avi_header.dwStreams << endl
        << "    dwSuggestedBufferSize: " << (int)main_avi_header.dwSuggestedBufferSize << endl
        << "    dwWidth:               " << (int)main_avi_header.dwWidth << endl
        << "    dwHeight:              " << (int)main_avi_header.dwHeight << endl;
        for (i = 0; i < 4; ++i)
            cout << "    dwReserved[" << i << "]:        " << (int)main_avi_header.dwReserved[i] << endl;

    } else if (entry.type == make_fourcc("strh")) {

        AVIStreamHeader	avi_stream_header;

        fail_if(lseek(fd, entry.offset, SEEK_SET) == (off_t)-1);
        fail_neg(read(fd, &avi_stream_header, sizeof(AVIStreamHeader)));

        cout << "    fccType:         '"
        << (char)((avi_stream_header.fccType >> 0) & 0x000000ff)
        << (char)((avi_stream_header.fccType >> 8) & 0x000000ff)
        << (char)((avi_stream_header.fccType >> 16) & 0x000000ff)
        << (char)((avi_stream_header.fccType >> 24) & 0x000000ff)
        << '\'' << endl
        <<      "    fccHandler:      '"
        << (char)((avi_stream_header.fccHandler >> 0) & 0x000000ff)
        << (char)((avi_stream_header.fccHandler >> 8) & 0x000000ff)
        << (char)((avi_stream_header.fccHandler >> 16) & 0x000000ff)
        << (char)((avi_stream_header.fccHandler >> 24) & 0x000000ff)
        << '\'' << endl
        << "    dwFlags:         " << (int)avi_stream_header.dwFlags << endl
        << "    wPriority:       " << (int)avi_stream_header.wPriority << endl
        << "    wLanguage:       " << (int)avi_stream_header.wLanguage << endl
        << "    dwInitialFrames: " << (int)avi_stream_header.dwInitialFrames << endl
        << "    dwScale:         " << (int)avi_stream_header.dwScale << endl
        << "    dwRate:          " << (int)avi_stream_header.dwRate << endl
        << "    dwLength:        " << (int)avi_stream_header.dwLength << endl
        << "    dwQuality:       " << (int)avi_stream_header.dwQuality << endl
        << "    dwSampleSize:    " << (int)avi_stream_header.dwSampleSize << endl;

    } else if (entry.type == make_fourcc("indx")) {

        int             i;
        AVISuperIndex   avi_super_index;

        fail_if(lseek(fd, entry.offset, SEEK_SET) == (off_t)-1);
        fail_neg(read(fd, &avi_super_index, sizeof(AVISuperIndex)));

        cout << "    wLongsPerEntry: " <<  (int)avi_super_index.wLongsPerEntry
        << endl
        << "    bIndexSubType:  " <<  (int)avi_super_index.bIndexSubType << endl
        << "    bIndexType:     " <<  (int)avi_super_index.bIndexType << endl
        << "    nEntriesInUse:  " <<  (int)avi_super_index.nEntriesInUse << endl
        << "    dwChunkId:      '"
        << (char)((avi_super_index.dwChunkId >> 0) & 0x000000ff)
        << (char)((avi_super_index.dwChunkId >> 8) & 0x000000ff)
        << (char)((avi_super_index.dwChunkId >> 16) & 0x000000ff)
        << (char)((avi_super_index.dwChunkId >> 24) & 0x000000ff)
        << '\'' << endl
        << "    dwReserved[0]:  " <<  (int)avi_super_index.dwReserved[0] << endl
        << "    dwReserved[1]:  " <<  (int)avi_super_index.dwReserved[1] << endl
        << "    dwReserved[2]:  " <<  (int)avi_super_index.dwReserved[2] << endl;
        for (i = 0; i < avi_super_index.nEntriesInUse; ++i) {
            cout << ' ' << setw(4) << setfill(' ') << i
            << ": qwOffset    : 0x" << setw(12) << setfill('0') << hex << avi_super_index.aIndex[i].qwOffset << endl
            << "       dwSize      : 0x" << setw(8) << avi_super_index.aIndex[i].dwSize << endl
            << "       dwDuration  : " << dec << avi_super_index.aIndex[i].dwDuration << endl;
        }
    }

    /* This is the Standard Index. It is an array of offsets and
       sizes relative to some start offset. */

    else if ((entry.type == make_fourcc("ix00")) || (entry.type == make_fourcc("ix01"))) {

        int             i;
        AVIStdIndex     avi_std_index;

        fail_if(lseek(fd, entry.offset, SEEK_SET) == (off_t)-1);
        fail_neg(read(fd, &avi_std_index, sizeof(AVIStdIndex)));

        cout << "    wLongsPerEntry: " << (int)avi_std_index.wLongsPerEntry
        << endl
        << "    bIndexSubType:  " << (int)avi_std_index.bIndexSubType << endl
        << "    bIndexType:     " << (int)avi_std_index.bIndexType << endl
        << "    nEntriesInUse:  " << (int)avi_std_index.nEntriesInUse << endl
        << "    dwChunkId:      '"
        << (char)((avi_std_index.dwChunkId >> 0) & 0x000000ff)
        << (char)((avi_std_index.dwChunkId >> 8) & 0x000000ff)
        << (char)((avi_std_index.dwChunkId >> 16) & 0x000000ff)
        << (char)((avi_std_index.dwChunkId >> 24) & 0x000000ff)
        << '\'' << endl
        << "    qwBaseOffset:   0x" << setw(12) << hex << avi_std_index.qwBaseOffset << endl
        << "    dwReserved:     " << dec << (int)avi_std_index.dwReserved << endl;
        for (i = 0; i < avi_std_index.nEntriesInUse; ++i) {
            cout << ' ' << setw(4) << setfill(' ') << i
            << ": dwOffset    : 0x" << setw(8) << setfill('0') << hex << avi_std_index.aIndex[i].dwOffset
            << " (0x" << setw(12) << avi_std_index.qwBaseOffset + avi_std_index.aIndex[i].dwOffset << ')' << endl
            << "       dwSize      : 0x" << setw(8) << avi_std_index.aIndex[i].dwSize << dec << endl;
        }

    } else if (entry.type == make_fourcc("idx1")) {

        int i;
        int numEntries = entry.length / sizeof(int) / 4;
        int *idx1 = new int[numEntries * 4];
        // FOURCC movi_list = FindDirectoryEntry(make_fourcc("movi"));

        fail_if(lseek(fd, entry.offset, SEEK_SET) == (off_t)-1);
        fail_neg(read(fd, idx1, entry.length));

        for (i = 0; i < numEntries; ++i) {

            cout << ' ' << setw(4) << setfill(' ') << i << setfill('0') << ": dwChunkId : '"
            << (char)((idx1[i * 4 + 0] >> 0) & 0x000000ff)
            << (char)((idx1[i * 4 + 0] >> 8) & 0x000000ff)
            << (char)((idx1[i * 4 + 0] >> 16) & 0x000000ff)
            << (char)((idx1[i * 4 + 0] >> 24) & 0x000000ff)
            << '\'' << endl
            << "       dwType    : 0x" << setw(8) << hex << idx1[i * 4 + 1] << endl
            << "       dwOffset  : 0x" << setw(8) << idx1[i * 4 + 2] << endl
            // << " (0x" << setw(8) << idx1[i * 4 + 2] + GetDirectoryEntry(movi_list).offset << ')' << endl
            << "       dwSize    : 0x" << setw(8) << idx1[i * 4 + 3] << dec << endl;
        }

        delete[] idx1;
    }
}


/** If this is not a movi list, read its contents
 
*/

void AVIFile::ParseList(int parent)
{
    FOURCC      type;
    FOURCC      name;
    int         length;
    int         list;
    off_t       pos;
    off_t       listEnd;

    /* Read in the chunk header (type and length). */
    fail_neg(read(fd, &type, sizeof(type)));
    fail_neg(read(fd, &length, sizeof(length)));
    if (length & 1)
        length++;

    /* The contents of the list starts here. Obtain its offset. The list
       name (4 bytes) is already part of the contents). */

    pos = lseek(fd, 0, SEEK_CUR);
    fail_if(pos == (off_t)-1);
    fail_neg(read(fd, &name, sizeof(name)));

    /* if we encounter a movi list, do not read it. It takes too much time
       and we don't need it anyway. */

    if (name != make_fourcc("movi")) {
//    if (1) {

        /* Add an entry for this list. */
        list = AddDirectoryEntry(type, name, sizeof(name), parent);

        /* Read in any chunks contained in this list. This list is the
           parent for all chunks it contains. */

        listEnd = pos + length;
        while (pos < listEnd) {
            ParseChunk(list);
            pos = lseek(fd, 0, SEEK_CUR);
            fail_if(pos == (off_t)-1);
        }
    } else {
        /* Add an entry for this list. */

        movi_list = AddDirectoryEntry(type, name, length, parent);

        pos = lseek(fd, length - 4, SEEK_CUR);
        fail_if(pos == (off_t)-1);
    }
}


void AVIFile::ParseRIFF()
{
    RIFFFile::ParseRIFF();

    avih_chunk = FindDirectoryEntry(make_fourcc("avih"));
    if (avih_chunk != -1)
        ReadChunk(avih_chunk, (void*)&mainHdr);
}


void AVIFile::ReadIndex()
{
    /* note: how do we find the second indx chunk ??? */
    indx_chunk[0] = FindDirectoryEntry(make_fourcc("indx"));
    if (indx_chunk[0] != -1) {
        ReadChunk(indx_chunk[0], (void*)indx[0]);
        index_type = AVI_LARGE_INDEX;
        return;
    }
    idx1_chunk = FindDirectoryEntry(make_fourcc("idx1"));
    if (idx1_chunk != -1) {
        ReadChunk(idx1_chunk, (void*)idx1);
        idx1->nEntriesInUse = GetDirectoryEntry(idx1_chunk).length / 16;
        index_type = AVI_SMALL_INDEX;
        return;
    }
}


void AVIFile::FlushIndx(int stream)
{
    FOURCC      type;
    FOURCC      name;
    off_t       length;
    off_t       offset;
    int         parent;
    int         i;

    /* Write out the previous index. When this function is
       entered for the first time, there is no index to
       write.  Note: this may be an expensive operation
       because of a time consuming seek to the former file
       position. */

    if (ix_chunk[stream] != -1)
        WriteChunk(ix_chunk[stream], ix[stream]);

    /* make a new ix chunk. */

    if (stream == 0)
        type = make_fourcc("ix00");
    else
        type = make_fourcc("ix01");
    ix_chunk[stream] = AddDirectoryEntry(type, 0, sizeof(AVIStdIndex), movi_list);
    GetDirectoryEntry(ix_chunk[stream], type, name, length, offset, parent);

    /* fill out all required fields. The offsets in the
       array are relative to qwBaseOffset, so fill in the
       offset to the next free location in the file
       there. */

    ix[stream]->wLongsPerEntry = 2;
    ix[stream]->bIndexSubType = 0;
    ix[stream]->bIndexType = AVI_INDEX_OF_CHUNKS;
    ix[stream]->nEntriesInUse = 0;
    ix[stream]->dwChunkId = indx[stream]->dwChunkId;
    ix[stream]->qwBaseOffset = offset + length;
    ix[stream]->dwReserved = 0;

    for (i = 0; i < IX00_INDEX_SIZE; ++i) {
        ix[stream]->aIndex[i].dwOffset = 0;
        ix[stream]->aIndex[i].dwSize = 0;
    }

    /* add a reference to this new index in our super
       index. */

    i = indx[stream]->nEntriesInUse++;
    indx[stream]->aIndex[i].qwOffset = offset - RIFF_HEADERSIZE;
    indx[stream]->aIndex[i].dwSize = length + RIFF_HEADERSIZE;
    indx[stream]->aIndex[i].dwDuration = 0;
}


void AVIFile::UpdateIndx(int stream, int chunk)
{
    FOURCC      type;
    FOURCC      name;
    off_t       length;
    off_t       offset;
    int         parent;
    int         i;

    /* update the appropiate entry in the super index. It reflects
       the number of frames in the referenced index. */

    i = indx[stream]->nEntriesInUse - 1;
    ++indx[stream]->aIndex[i].dwDuration;

    /* update the standard index. Calculate the file position of
       the new frame. */

    GetDirectoryEntry(chunk, type, name, length, offset, parent);

    indx[stream]->dwChunkId = type;
    i = ix[stream]->nEntriesInUse++;
    ix[stream]->aIndex[i].dwOffset = offset - ix[stream]->qwBaseOffset;
    ix[stream]->aIndex[i].dwSize = length;
}


void AVIFile::UpdateIdx1(int chunk, int flags)
{
	if (idx1->nEntriesInUse < 20000) {
    FOURCC      type;
    FOURCC      name;
    off_t       length;
    off_t       offset;
    int         parent;

    GetDirectoryEntry(chunk, type, name, length, offset, parent);

    idx1->aIndex[idx1->nEntriesInUse].dwChunkId = type;
    idx1->aIndex[idx1->nEntriesInUse].dwFlags = flags;
    idx1->aIndex[idx1->nEntriesInUse].dwOffset = offset - GetDirectoryEntry(movi_list).offset - RIFF_HEADERSIZE;
    idx1->aIndex[idx1->nEntriesInUse].dwSize = length;
    idx1->nEntriesInUse++;
	}
}


AVI1File::AVI1File() : AVIFile(), odml_list( -1), dmlh_chunk( -1)
{}


AVI1File::~AVI1File()
{}


/* Initialize the AVI structure to its initial state, either for PAL
   or NTSC format */

void AVI1File::Init(int format, int sampleFrequency, int indexType)
{
    int i;

    assert((format == AVI_PAL) || (format == AVI_NTSC));

    AVIFile::Init(format, sampleFrequency, indexType);

    switch (format) {
    case AVI_PAL:
        streamHdr[0].dwScale = 400000;
        streamHdr[0].dwSuggestedBufferSize = 144008;

        /* I do not know what these values mean, I copied them
           from an existing PAL AVI file */

        /* initialize the 'strf' chunk */

        dvinfo.dwDVAAuxSrc = 0xd1e030d0;
        dvinfo.dwDVAAuxCtl = 0xffa0cf3f;
        dvinfo.dwDVAAuxSrc1 = 0xd1e03fd0;
        dvinfo.dwDVAAuxCtl1 = 0xffa0cf3f;
        dvinfo.dwDVVAuxSrc = 0xff20ffff;
        dvinfo.dwDVVAuxCtl = 0xfffdc83f;
        dvinfo.dwDVReserved[0] = 0;
        dvinfo.dwDVReserved[1] = 0;
        break;

    case AVI_NTSC:
        streamHdr[0].dwScale = 333666;
        streamHdr[0].dwSuggestedBufferSize = 120008;

        /* I do not know what these values mean, I copied them from an existing NTSC AVI file */

        /* initialize the 'strf' chunk */

        dvinfo.dwDVAAuxSrc = 0xc0c000c0;
        dvinfo.dwDVAAuxCtl = 0xffa0cf3f;
        dvinfo.dwDVAAuxSrc1 = 0xc0c001c0;
        dvinfo.dwDVAAuxCtl1 = 0xffa0cf3f;
        dvinfo.dwDVVAuxSrc = 0xff80ffff;
        dvinfo.dwDVVAuxCtl = 0xfffcc83f;
        dvinfo.dwDVReserved[0] = 0;
        dvinfo.dwDVReserved[1] = 0;
        break;

    default: /* no default allowed */
        assert(0);
        break;
    }

    indx[0]->dwChunkId = make_fourcc("00__");
    indx[1]->dwChunkId = make_fourcc("01__");

    /* Initialize the 'strh' chunk */

    streamHdr[0].fccType = make_fourcc("iavs");
    streamHdr[0].fccHandler = make_fourcc("dvsd");
    streamHdr[0].dwFlags = 0;
    streamHdr[0].wPriority = 0;
    streamHdr[0].wLanguage = 0;
    streamHdr[0].dwInitialFrames = 0;
    streamHdr[0].dwRate = 10000000;
    streamHdr[0].dwStart = 0;
    streamHdr[0].dwLength = 0;
    streamHdr[0].dwQuality = 0;
    streamHdr[0].dwSampleSize = 0;
    streamHdr[0].rcFrame.top = 0;
    streamHdr[0].rcFrame.bottom = 0;
    streamHdr[0].rcFrame.left = 0;
    streamHdr[0].rcFrame.right = 0;

    /* Initialize the 'dmlh' chunk. I have no clue what this means
       though */

    for (i = 0; i < 62; ++i)
        (dmlh)[i] = 0;
    (dmlh)[0] = -1;            /* frame count + 1? */

    /* This is a simple directory structure setup. For details see the
       "OpenDML AVI File Format Extensions" document.
       
       An AVI file contains basically two types of objects, a
       "chunk" and a "list" object. The list object contains any
       number of chunks. Since a list is also a chunk, it is
       possible to create a hierarchical "list of lists"
       structure.

       Every AVI file starts with a "RIFF" object, which is a list
       of several other required objects. The actual DV data is
       contained in a "movi" list, each frame is in its own chunk.

       Old AVI files (pre OpenDML V. 1.02) contain only one RIFF
       chunk of less than 1 GByte size per file. The current
       format which allow for almost arbitrary sizes can contain
       several RIFF chunks of less than 1 GByte size. Old software
       however would only deal with the first RIFF chunk.

       Note that the first entry (g_file_list) isnt actually part
       of the AVI file. I use this (pseudo-) directory entry to
       keep track of the RIFF chunks and their positions in the
       AVI file. */

    /* Create the container directory entry */

    file_list = AddDirectoryEntry(make_fourcc("FILE"), make_fourcc("FILE"), 0, RIFF_NO_PARENT);

    /* Create a basic directory structure. Only chunks defined from here on will be written to the AVI file. */

    riff_list = AddDirectoryEntry(make_fourcc("RIFF"), make_fourcc("AVI "), RIFF_LISTSIZE, file_list);
    hdrl_list = AddDirectoryEntry(make_fourcc("LIST"), make_fourcc("hdrl"), RIFF_LISTSIZE, riff_list);
    avih_chunk = AddDirectoryEntry(make_fourcc("avih"), 0, sizeof(MainAVIHeader), hdrl_list);
    strl_list[0] = AddDirectoryEntry(make_fourcc("LIST"), make_fourcc("strl"), RIFF_LISTSIZE, hdrl_list);
    strh_chunk[0] = AddDirectoryEntry(make_fourcc("strh"), 0, sizeof(AVIStreamHeader), strl_list[0]);
    strf_chunk[0] = AddDirectoryEntry(make_fourcc("strf"), 0, sizeof(dvinfo), strl_list[0]);
    if (index_type & AVI_LARGE_INDEX)
        indx_chunk[0] = AddDirectoryEntry(make_fourcc("indx"), 0, sizeof(AVISuperIndex), strl_list[0]);
    odml_list = AddDirectoryEntry(make_fourcc("LIST"), make_fourcc("odml"), RIFF_LISTSIZE, hdrl_list);
    dmlh_chunk = AddDirectoryEntry(make_fourcc("dmlh"), 0, 0x00f8, odml_list);
    junk_chunk = AddDirectoryEntry(make_fourcc("JUNK"), 0, 0x0014, riff_list);
    movi_list = AddDirectoryEntry(make_fourcc("LIST"), make_fourcc("movi"), RIFF_LISTSIZE, riff_list);

    /* The ix00 chunk will be added dynamically to the movi_list in avi_write_frame
              as needed */

    ix_chunk[0] = -1;
}


/* Write a DV video frame. This is somewhat complex... */

void AVI1File::WriteFrame(const Frame &frame)
{
    int         num_frames;
    int         frame_chunk;
    int         junk_chunk;
    int         num_blocks;
    FOURCC      type;
    FOURCC      name;
    off_t       length;
    off_t       offset;
    int         parent;

    /* Check if we need a new ix00 Standard Index. It has a
       capacity of IX00_INDEX_SIZE frames. Whenever we exceed that
       number, we need a new index. The new ix00 chunk is also
       part of the movi list. */

    if ((index_type & AVI_LARGE_INDEX) && (((streamHdr[0].dwLength - 0) % IX00_INDEX_SIZE) == 0))
        FlushIndx(0);

    /* Write the DV frame data.

       Make a new 00__ chunk for the new frame, write out the
       frame, then add a JUNK chunk which is sized such that we
       end up on a 512 bytes boundary. */

    frame_chunk = AddDirectoryEntry(make_fourcc("00__"), 0, frame.GetFrameSize(), movi_list);
    WriteChunk(frame_chunk, frame.data);
    num_blocks = frame.GetFrameSize() / PADDING_SIZE + 1;
    junk_chunk = AddDirectoryEntry(make_fourcc("JUNK"), 0, num_blocks * PADDING_SIZE - frame.GetFrameSize() - 2 * RIFF_HEADERSIZE, movi_list);
    WriteChunk(junk_chunk, g_zeroes);

    if (index_type & AVI_LARGE_INDEX)
        UpdateIndx(0, frame_chunk);
    if (index_type & AVI_SMALL_INDEX)
        UpdateIdx1(frame_chunk, 0x10);

    /* update some variables with the new frame count. */

    num_frames = ++mainHdr.dwTotalFrames;
    streamHdr[0].dwLength = num_frames;
    (dmlh)[0] = num_frames;

    /* Find out if the current riff list is close to 1 GByte in
       size. If so, start a new (extended) RIFF. The only allowed
       item in the new RIFF chunk is a movi list (with video
       frames and indexes as usual). */

    GetDirectoryEntry(riff_list, type, name, length, offset, parent);
    if (length > 0x3f000000) {
        riff_list = AddDirectoryEntry(make_fourcc("RIFF"), make_fourcc("AVIX"), RIFF_LISTSIZE, file_list);
        movi_list = AddDirectoryEntry(make_fourcc("LIST"), make_fourcc("movi"), RIFF_LISTSIZE, riff_list);
        FlushIndx(0);
    }
}


void AVI1File::WriteRIFF()
{

    WriteChunk(avih_chunk, (void*)&mainHdr);
    WriteChunk(strh_chunk[0], (void*)&streamHdr[0]);
    WriteChunk(strf_chunk[0], (void*)&dvinfo);
    WriteChunk(dmlh_chunk, (void*)&dmlh);

    if (index_type & AVI_LARGE_INDEX) {
        WriteChunk(indx_chunk[0], (void*)indx[0]);
        WriteChunk(ix_chunk[0], (void*)ix[0]);
    }
    
    if (index_type & AVI_SMALL_INDEX) {
        int idx1_chunk = AddDirectoryEntry(make_fourcc("idx1"), 0, idx1->nEntriesInUse * 16, riff_list);
        WriteChunk(idx1_chunk, (void*)idx1);
    }

    RIFFFile::WriteRIFF();
}


AVI2File::AVI2File() : AVIFile()
{}


AVI2File::~AVI2File()
{}


/* Initialize the AVI structure to its initial state, either for PAL
   or NTSC format */

void AVI2File::Init(int format, int sampleFrequency, int indexType)
{
    assert((format == AVI_PAL) || (format == AVI_NTSC));

    AVIFile::Init(format, sampleFrequency, indexType);

    switch (format) {

    case AVI_PAL:
        mainHdr.dwPaddingGranularity = 0;
        mainHdr.dwFlags = AVIF_HASINDEX | AVIF_TRUSTCKTYPE;
        mainHdr.dwStreams = 2;
        mainHdr.dwWidth = 720;
        mainHdr.dwHeight = 576;
        mainHdr.dwReserved[0] = 40000;
        mainHdr.dwReserved[1] = 1000000;
        mainHdr.dwReserved[2] = 0;
        mainHdr.dwReserved[3] = 0;

        /* Initialize the 'strh' chunk */

        streamHdr[0].fccType = make_fourcc("vids");
        streamHdr[0].fccHandler = make_fourcc("dvsd");
        streamHdr[0].dwFlags = 0;
        streamHdr[0].wPriority = 0;
        streamHdr[0].wLanguage = 0;
        streamHdr[0].dwInitialFrames = 0;
        streamHdr[0].dwScale = 40000;
        streamHdr[0].dwRate = 1000000;
        streamHdr[0].dwStart = 0;
        streamHdr[0].dwLength = 0;
        streamHdr[0].dwSuggestedBufferSize = 0;
        streamHdr[0].dwQuality = -1;
        streamHdr[0].dwSampleSize = 0;
        streamHdr[0].rcFrame.top = 0;
        streamHdr[0].rcFrame.bottom = 0;
        streamHdr[0].rcFrame.left = 0;
        streamHdr[0].rcFrame.right = 0;

        bitmapinfo.dwSize = sizeof(bitmapinfo);
        bitmapinfo.biWidth = 720;
        bitmapinfo.biHeight = 576;
        bitmapinfo.biPlanes = 1;
        bitmapinfo.biBitCount = 24;
        bitmapinfo.biCompression = make_fourcc("dvsd");
        bitmapinfo.biSizeImage = bitmapinfo.biWidth * bitmapinfo.biHeight * 3;
        bitmapinfo.biXPelsPerMeter = 0;
        bitmapinfo.biYPelsPerMeter = 0;
        bitmapinfo.biClrUsed = 0;
        bitmapinfo.biClrImportant = 0;

        streamHdr[1].fccType = make_fourcc("auds");
        streamHdr[1].fccHandler = 0;
        streamHdr[1].dwFlags = 0;
        streamHdr[1].wPriority = 0;
        streamHdr[1].wLanguage = 0;
        streamHdr[1].dwInitialFrames = 0;
        streamHdr[1].dwScale = 1;
        streamHdr[1].dwRate = sampleFrequency;
        streamHdr[1].dwStart = 0;
        streamHdr[1].dwLength = 0;
        streamHdr[1].dwSuggestedBufferSize = 0;
        streamHdr[1].dwQuality = -1;
        streamHdr[1].dwSampleSize = 2;
        streamHdr[1].rcFrame.top = 0;
        streamHdr[1].rcFrame.bottom = 0;
        streamHdr[1].rcFrame.left = 0;
        streamHdr[1].rcFrame.right = 0;

        break;

    case AVI_NTSC:
        mainHdr.dwMaxBytesPerSec = 4147200;
        mainHdr.dwPaddingGranularity = 0;
        mainHdr.dwFlags = AVIF_HASINDEX | AVIF_TRUSTCKTYPE;
        mainHdr.dwTotalFrames = 0;
        mainHdr.dwInitialFrames = 0;
        mainHdr.dwStreams = 2;
        mainHdr.dwWidth = 720;
        mainHdr.dwHeight = 480;
        mainHdr.dwReserved[0] = 33366;
        mainHdr.dwReserved[1] = 1000000;
        mainHdr.dwReserved[2] = 0;
        mainHdr.dwReserved[3] = 0;

        /* Initialize the 'strh' chunk */

        streamHdr[0].fccType = make_fourcc("vids");
        streamHdr[0].fccHandler = make_fourcc("dvsd");
        streamHdr[0].dwFlags = 0;
        streamHdr[0].wPriority = 0;
        streamHdr[0].wLanguage = 0;
        streamHdr[0].dwInitialFrames = 0;
        streamHdr[0].dwScale = 33366;
        streamHdr[0].dwRate = 1000000;
        streamHdr[0].dwStart = 0;
        streamHdr[0].dwLength = 0;
        streamHdr[0].dwSuggestedBufferSize = 0;
        streamHdr[0].dwQuality = -1;
        streamHdr[0].dwSampleSize = 0;
        streamHdr[0].rcFrame.top = 0;
        streamHdr[0].rcFrame.bottom = 0;
        streamHdr[0].rcFrame.left = 0;
        streamHdr[0].rcFrame.right = 0;

        bitmapinfo.dwSize = sizeof(bitmapinfo);
        bitmapinfo.biWidth = 720;
        bitmapinfo.biHeight = 480;
        bitmapinfo.biPlanes = 1;
        bitmapinfo.biBitCount = 24;
        bitmapinfo.biCompression = make_fourcc("dvsd");
        bitmapinfo.biSizeImage = bitmapinfo.biWidth * bitmapinfo.biHeight * 3;
        bitmapinfo.biXPelsPerMeter = 0;
        bitmapinfo.biYPelsPerMeter = 0;
        bitmapinfo.biClrUsed = 0;
        bitmapinfo.biClrImportant = 0;

        streamHdr[1].fccType = make_fourcc("auds");
        streamHdr[1].fccHandler = 0;
        streamHdr[1].dwFlags = 0;
        streamHdr[1].wPriority = 0;
        streamHdr[1].wLanguage = 0;
        streamHdr[1].dwInitialFrames = 0;
        streamHdr[1].dwScale = 1;
        streamHdr[1].dwRate = sampleFrequency;
        streamHdr[1].dwStart = 0;
        streamHdr[1].dwLength = 0;
        streamHdr[1].dwSuggestedBufferSize = 0;
        streamHdr[1].dwQuality = -1;
        streamHdr[1].dwSampleSize = 2;
        streamHdr[1].rcFrame.top = 0;
        streamHdr[1].rcFrame.bottom = 0;
        streamHdr[1].rcFrame.left = 0;
        streamHdr[1].rcFrame.right = 0;

        break;
    }
    waveformatex.wFormatTag = 1;
    waveformatex.nChannels = 2;
    waveformatex.nSamplesPerSec = sampleFrequency;
    waveformatex.nAvgBytesPerSec = sampleFrequency * 4;
    waveformatex.nBlockAlign = 4;
    waveformatex.wBitsPerSample = 16;

    file_list = AddDirectoryEntry(make_fourcc("FILE"), make_fourcc("FILE"), 0, RIFF_NO_PARENT);

    /* Create a basic directory structure. Only chunks defined from here on will be written to the AVI file. */

    riff_list = AddDirectoryEntry(make_fourcc("RIFF"), make_fourcc("AVI "), RIFF_LISTSIZE, file_list);
    hdrl_list = AddDirectoryEntry(make_fourcc("LIST"), make_fourcc("hdrl"), RIFF_LISTSIZE, riff_list);
    avih_chunk = AddDirectoryEntry(make_fourcc("avih"), 0, sizeof(MainAVIHeader), hdrl_list);

    strl_list[0] = AddDirectoryEntry(make_fourcc("LIST"), make_fourcc("strl"), RIFF_LISTSIZE, hdrl_list);
    strh_chunk[0] = AddDirectoryEntry(make_fourcc("strh"), 0, sizeof(AVIStreamHeader), strl_list[0]);
    strf_chunk[0] = AddDirectoryEntry(make_fourcc("strf"), 0, sizeof(BITMAPINFOHEADER), strl_list[0]);
    if (index_type & AVI_LARGE_INDEX) {
        indx_chunk[0] = AddDirectoryEntry(make_fourcc("indx"), 0, sizeof(AVISuperIndex), strl_list[0]);
        ix_chunk[0] = -1;
        indx[0]->dwChunkId = make_fourcc("00__");
    }

    strl_list[1] = AddDirectoryEntry(make_fourcc("LIST"), make_fourcc("strl"), RIFF_LISTSIZE, hdrl_list);
    strh_chunk[1] = AddDirectoryEntry(make_fourcc("strh"), 0, sizeof(AVIStreamHeader), strl_list[1]);
    strf_chunk[1] = AddDirectoryEntry(make_fourcc("strf"), 0, sizeof(WAVEFORMATEX), strl_list[1]);
    if (index_type & AVI_LARGE_INDEX) {
        indx_chunk[1] = AddDirectoryEntry(make_fourcc("indx"), 0, sizeof(AVISuperIndex), strl_list[1]);
        ix_chunk[1] = -1;
        indx[1]->dwChunkId = make_fourcc("01__");
    }

    movi_list = AddDirectoryEntry(make_fourcc("LIST"), make_fourcc("movi"), RIFF_LISTSIZE, riff_list);
}


void AVI2File::WriteRIFF()
{
    WriteChunk(avih_chunk, (void*)&mainHdr);
    WriteChunk(strh_chunk[0], (void*)&streamHdr[0]);
    WriteChunk(strf_chunk[0], (void*)&bitmapinfo);
    if (index_type & AVI_LARGE_INDEX) {
        WriteChunk(indx_chunk[0], (void*)indx[0]);
        WriteChunk(ix_chunk[0], (void*)ix[0]);
    }
    WriteChunk(strh_chunk[1], (void*)&streamHdr[1]);
    WriteChunk(strf_chunk[1], (void*)&waveformatex);
    if (index_type & AVI_LARGE_INDEX) {
        WriteChunk(indx_chunk[1], (void*)indx[1]);
        WriteChunk(ix_chunk[1], (void*)ix[1]);
    }
    if (index_type & AVI_SMALL_INDEX) {
        int idx1_chunk = AddDirectoryEntry(make_fourcc("idx1"), 0, idx1->nEntriesInUse * 16, riff_list);
        WriteChunk(idx1_chunk, (void*)idx1);
    }

    RIFFFile::WriteRIFF();
}


/** Write a DV video frame
 
    \param frame the frame to write
*/

void AVI2File::WriteFrame(const Frame &frame)
{
    int         num_frames;
    int         audio_chunk;
    int         frame_chunk;
    char        soundbuf[20000];
    int		audio_size;

    /* Check if we need a new ix00 Standard Index. It has a
       capacity of IX00_INDEX_SIZE frames. Whenever we exceed that
       number, we need a new index. The new ix00 chunk is also
       part of the movi list. */

    if ((index_type & AVI_LARGE_INDEX) && (((streamHdr[0].dwLength - 0) % IX00_INDEX_SIZE) == 0)) {
        FlushIndx(0);
        FlushIndx(1);
    }

    /* Write audio data if we have it */

    audio_size = frame.ExtractAudio(soundbuf);
    if (audio_size > 0) {
        audio_chunk = AddDirectoryEntry(make_fourcc("01wb"), 0, audio_size, movi_list);
        WriteChunk(audio_chunk, soundbuf);
        if (index_type & AVI_LARGE_INDEX)
            UpdateIndx(1, audio_chunk);
        if (index_type & AVI_SMALL_INDEX)
            UpdateIdx1(audio_chunk, 0x00);
        streamHdr[1].dwLength += audio_size;
    }

    /* Write video data */

    frame_chunk = AddDirectoryEntry(make_fourcc("00dc"), 0, frame.GetFrameSize(), movi_list);
    WriteChunk(frame_chunk, frame.data);
    if (index_type & AVI_LARGE_INDEX)
        UpdateIndx(0, frame_chunk);
    if (index_type & AVI_SMALL_INDEX)
        UpdateIdx1(frame_chunk, 0x10);

    /* update some variables with the new frame count. */

    num_frames = ++mainHdr.dwTotalFrames;
    streamHdr[0].dwLength = num_frames;
    mainHdr.dwReserved[3] = num_frames;
}
