/*
 * filehandler.cc -- saving DV data into different file formats
 * Copyright (C) 2000 Arne Schirmacher <arne@schirmacher.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>

#include <string>
#include <iostream>
#include <strstream>
#include <iomanip>

using std::cerr;
using std::cout;
using std::endl;
using std::strstream;
using std::setw;
using std::setfill;

#include "error.h"
#include "riff.h"
#include "avi.h"
#include "frame.h"
#include "filehandler.h"

FileTracker *FileTracker::instance = NULL;

FileTracker::FileTracker( ) : mode( CAPTURE_MOVIE_APPEND )
{
	cout << ">> Constructing File Capture tracker" << endl;
}

FileTracker::~FileTracker( )
{
	cout << ">> Destroying File Capture tracker" << endl;
}

FileTracker &FileTracker::GetInstance( )
{
	if ( instance == NULL )
		instance = new FileTracker();

	return *instance;
}

void FileTracker::SetMode( FileCaptureMode mode )
{
	this->mode = mode;
}

FileCaptureMode FileTracker::GetMode( )
{
	return this->mode;
}

char *FileTracker::Get( int index )
{
	return list[ index ];
}

void FileTracker::Add( const char *file )
{
	if ( this->mode != CAPTURE_IGNORE ) 
	{
		cout << ">>>> Registering " << file << " with the tracker" << endl;
		list.push_back( strdup( file ) );
	}
}

unsigned int FileTracker::Size( )
{
	return list.size();
}

void FileTracker::Clear( )
{
	while ( Size() > 0 )
	{
		free( list[ Size() - 1 ] );
		list.pop_back( );
	}
	this->mode = CAPTURE_MOVIE_APPEND;
}

FileHandler::FileHandler() : done(false), autoSplit(false), maxFrameCount(999999), framesWritten(0)
{
    /* empty body */
}


FileHandler::~FileHandler()
{
    /* empty body */
}


bool FileHandler::GetAutoSplit() const
{
    return autoSplit;
}


bool FileHandler::GetTimeStamp() const
{
    return timeStamp;
}


string FileHandler::GetBaseName() const
{
    return base;
}


string FileHandler::GetExtension() const
{
    return extension;
}


int FileHandler::GetMaxFrameCount() const
{
    return maxFrameCount;
}

off_t FileHandler::GetMaxFileSize() const
{
	return maxFileSize;
}


void FileHandler::SetAutoSplit(bool flag)
{
    autoSplit = flag;
}


void FileHandler::SetTimeStamp(bool flag)
{
    timeStamp = flag;
}


void FileHandler::SetBaseName(const string& s)
{
    base = s;
}


void FileHandler::SetMaxFrameCount(int count)
{
    assert(count >= 0);
    maxFrameCount = count;
}


void FileHandler::SetEveryNthFrame(int every)
{
    assert (every > 0);

    everyNthFrame = every;
}


void FileHandler::SetMaxFileSize(off_t size)
{
	assert (size >= 0);
	maxFileSize = size;
}


void FileHandler::SetSampleFrame(const Frame& sample)
{
    /* empty body */
}


bool FileHandler::Done()
{
    return done;
}


int FileHandler::WriteFrame(const Frame& frame)
{
    /* If the user wants autosplit, start a new file if a
       new recording is detected. */

    if (FileIsOpen() && GetAutoSplit() == true && frame.IsNewRecording()) {
        Close();
    }

    if (FileIsOpen() == false) {

        string          filename;
        static int      counter = 0;

        if (GetTimeStamp() == true) {
        	strstream	sb, sb2;
			struct tm	date;
			string		recDate;
		
			if (frame.GetRecordingDate(date) == true) {
				sb << setfill('0')
				<< setw(4) << date.tm_year + 1900 << '.'
				<< setw(2) << date.tm_mon + 1 << '.'
				<< setw(2) << date.tm_mday << '_'
				<< setw(2) << date.tm_hour << '-'
				<< setw(2) << date.tm_min << '-'
				<< setw(2) << date.tm_sec;
				sb >> recDate;
			} else {
				recDate = "0000.00.00_00-00-00";
			}
            sb2 << GetBaseName() << recDate << GetExtension();
			sb2 >> filename;
			cout << ">>> Trying " << filename << endl;
        } else {
			struct stat stats;
			do {
        		strstream       sb;
            	sb << GetBaseName() << setfill('0') << setw(3) << ++ counter << GetExtension();
        		sb >> filename;
				cout << ">>> Trying " << filename << endl;
			}
			while ( stat( filename.c_str(), &stats ) == 0 );
        }
        fail_if(Create(filename) == false)
        ;
        framesWritten = 0;
        framesToSkip = 0;
    }

    /* write frame */

    if (framesToSkip == 0) {
      if (0 > Write(frame)) {
	cerr << "Error writing frame!" << endl;
      }
        framesToSkip = everyNthFrame;
        ++framesWritten;
    }
    framesToSkip--;
	
    /* If the frame count is exceeded, close the current file.
       If the autosplit flag is set, a new file will be created in the next iteration.
       If the flag is not set, we are done. */

    if ( (GetMaxFrameCount() > 0) &&
		 (framesWritten >= GetMaxFrameCount()) ) {
        Close();
        if (GetAutoSplit() == false)
            done = true;
    }
	
    /* If the file size could be exceeded by another frame, close the current file.
       If the autosplit flag is set, a new file will be created on the next iteration.
       If the flag is not set, we are done. */
	/* not exact, but should be good enough to prevent going over. */
	if (FileIsOpen()) {
		AudioInfo info;
		frame.GetAudioInfo(info);
		if ( (GetFileSize() > 0) && 
			 (GetFileSize() + frame.GetFrameSize() + info.samples*4 + 12)
				>= GetMaxFileSize()) {                     // 12 = sizeof chunk metadata
			Close();
			if (GetAutoSplit() == false)
				done = true;
		}
	}

    return 0;
}


RawHandler::RawHandler(): fd( -1)
{
    extension = ".dv";
}


RawHandler::~RawHandler()
{
    /* empty body */
}


bool RawHandler::FileIsOpen()
{
    return fd != -1;
}


bool RawHandler::Create(const string& filename)
{
    fd = open(filename.c_str(), O_CREAT | O_TRUNC | O_RDWR, 0644);
    if (fd != -1) FileTracker::GetInstance().Add(filename.c_str());
    return fd != -1 ? true : false;
}


int RawHandler::Write(const Frame& frame)
{
    return write(fd, frame.data, frame.GetFrameSize());
}


int RawHandler::Close()
{
    if (fd != -1) {
        close(fd);
        fd = -1;
    }
    return 0;
}


off_t RawHandler::GetFileSize()
{
	struct stat file_status;
	fstat(fd, &file_status);
	return file_status.st_size;
}

int RawHandler::GetTotalFrames()
{
	return GetFileSize() / (480 * numBlocks);
}


bool RawHandler::Open(const char *s)
{
	unsigned char data[4];
	assert(fd == -1);
    fd = open(s, O_RDWR | O_NONBLOCK);
	if ( fd < 0 ) return false;
	if ( read(fd, data, 4) < 0 )
		return false;
	lseek( fd, 0, SEEK_SET);
	numBlocks = ((data[3] & 0x80) == 0) ? 250 : 300;
    return true;
	
}

int RawHandler::GetFrame(Frame &frame, int frameNum)
{
	assert(fd != -1);
	int size = 480 * numBlocks;
	if ( frameNum < 0 )
		return -1;
	off_t offset = ((off_t)frameNum * (off_t)size);
    fail_if(lseek(fd, offset, SEEK_SET) == (off_t)-1);
    if ( read(fd, frame.data, size) > 0 )
		return 0;
	else
		return -1;
}


/***************************************************************************/


AVIHandler::AVIHandler(int format): avi(NULL), aviFormat(format), isOpenDML(false)
{
    extension = ".avi";
}


AVIHandler::~AVIHandler()
{
    /* empty body */
}


void AVIHandler::SetSampleFrame(const Frame& sample)
{
    sample.GetAudioInfo(audioInfo);
    sample.GetVideoInfo(videoInfo);
}


bool AVIHandler::FileIsOpen()
{
    return avi != NULL;
}


bool AVIHandler::Create(const string& filename)
{
    assert(avi == NULL);

    switch (aviFormat) {

    case AVI_DV1_FORMAT:
        fail_null(avi = new AVI1File);
        fail_if(avi->Create(filename.c_str()) == false);
		avi->Init(videoInfo.isPAL ? AVI_PAL : AVI_NTSC, audioInfo.frequency, 
			AVI_LARGE_INDEX ); 
        break;

    case AVI_DV2_FORMAT:
        fail_null(avi = new AVI2File);
        fail_if(avi->Create(filename.c_str()) == false);
		if (GetOpenDML())
			avi->Init(videoInfo.isPAL ? AVI_PAL : AVI_NTSC, audioInfo.frequency, 
				(AVI_SMALL_INDEX | AVI_LARGE_INDEX) ); 
		else
			avi->Init(videoInfo.isPAL ? AVI_PAL : AVI_NTSC, audioInfo.frequency, 
				(AVI_SMALL_INDEX) ); 
        break;

    default:
        assert(aviFormat == AVI_DV1_FORMAT || aviFormat == AVI_DV2_FORMAT);
    }


    FileTracker::GetInstance().Add(filename.c_str());
    return avi != NULL ? true : false;
}


int AVIHandler::Write(const Frame& frame)
{
    assert(avi != NULL);

    avi->WriteFrame(frame);
    return 0;
}


int AVIHandler::Close()
{
    if (avi != NULL) {
        avi->WriteRIFF();
        delete avi;
        avi = NULL;
    }
    return 0;
}

off_t AVIHandler::GetFileSize()
{
	return avi->GetFileSize();
}

int AVIHandler::GetTotalFrames()
{
	return avi->GetTotalFrames();
}


bool AVIHandler::Open(const char *s)
{
	assert(avi == NULL);
	fail_null(avi = new AVI1File);
	if (avi->Open(s)) {
		avi->ParseRIFF();
		avi->ReadIndex();
		return true;
	} else
		return false;
	
}

int AVIHandler::GetFrame(Frame &frame, int frameNum)
{
	return avi->GetDVFrame(frame, frameNum);
}


void AVIHandler::SetOpenDML(bool flag)
{
	isOpenDML = flag;
}


bool AVIHandler::GetOpenDML() const
{
    return isOpenDML;
}


