/*
 *
 * Copyright (C) 2000 Arne Schirmacher <arne@schirmacher.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "framedisplayer.h"
#include "preferences.h"
#include <libdv/dv.h>
#include "frame.h"

#include <iostream>

#include <pthread.h>

using std::cout;
using std::cerr;
using std::endl;

FrameDisplayer::FrameDisplayer(): displayer(NULL), resampler( NULL ), audio_device_avail(FALSE)
{
    Preferences &prefs = Preferences::getInstance();
    if (prefs.enableAudio) audio_device = dv_oss_new();
    for(gint i=0; i < 4; i++)
        audio_buffers[i] = (gint16 *) malloc(2*DV_AUDIO_MAX_SAMPLES*sizeof(gint16));
}


FrameDisplayer::~FrameDisplayer()
{
	if ( displayer != NULL )
		delete displayer;

	if ( resampler != NULL )
		delete resampler;

    if (audio_device_avail) {
        dv_oss_close(audio_device);
        audio_device_avail = FALSE;
        audio_device = NULL;
    }
    for(gint i=0; i < 4; i++) free( audio_buffers[i]);
}

void FrameDisplayer::PutSound( Frame &frame)
{
    static Preferences &prefs = Preferences::getInstance();
	if (prefs.enableAudio && audio_device != NULL) {
		if ( !audio_device_avail &&
			( audio_sampling_rate = dv_oss_init(frame.decoder->audio, audio_device)) != 0 )
			audio_device_avail = TRUE;
		if ( audio_device_avail ) {
			if ( resampler == NULL )
				resampler = new FastAudioResample( audio_sampling_rate );
			resampler->Resample( frame );
			if ( resampler->size )
				dv_oss_player( audio_device, resampler->output, resampler->size );
		} else {
		  /* Audio is enabled, but init failed. 
		     Lets sleep.
		     This, of course, does not work well. */
		  struct timespec t;
		  t.tv_sec = 0;
		  // Rough estimate to normalise playback and allowing overhead of frame rendering
		  t.tv_nsec = frame.IsPAL() ? 2500000 : 2000000;
		  nanosleep( &t, NULL );
		  
		}
	}
}

void FrameDisplayer::Put(Frame &frame, GtkWidget *drawingarea, gboolean no_audio)
{
    static Preferences &prefs = Preferences::getInstance();

	if ( displayer == NULL ) 
		displayer = FindDisplayer::getDisplayer( drawingarea, frame );

	DisplayerInput input = DISPLAY_NONE;
	switch( displayer->format() ) {
		case DISPLAY_YUV:
			frame.ExtractYUV( pixels );
			input = DISPLAY_YUV;
			break;
		case DISPLAY_RGB:
			frame.ExtractRGB( pixels );
			input = DISPLAY_RGB;
			break;
	}

    if (!no_audio)  {
		PutSound(frame);
	}
    else if ( prefs.enableNoAudioDelay ) {
      struct timespec t;
		t.tv_sec = 0;
		// Rough estimate to normalise playback and allowing overhead of frame rendering
		t.tv_nsec = frame.IsPAL() ? 2500000 : 2000000;
		nanosleep( &t, NULL );
	}

	displayer->put( input, pixels, frame.GetWidth(), frame.GetHeight() );
}

