/*
 * page_export_1394.cc Notebook Firewire/AVI/Still Frame Export Page Object
 * Copyright (C) 2001 Dan Dennedy <dan@dennedy.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <iostream>
using std::cout;
using std::endl;

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <gnome.h>
#include <stdio.h>
#include <stdlib.h>

#include "page_export_1394.h"
#include "preferences.h"
#include "ieee1394io.h"
#include "message.h"

extern "C" {
#include "support.h"
#include "callbacks.h"
#include "interface.h"
#include "commands.h"
}

/** Constructor for page.

\param exportPage the export page object to bind to
\param common	common object to which this page belongs
*/

Export1394::Export1394(PageExport *_exportPage, KinoCommon *_common ) : 
  Export(_exportPage, _common), exportReader(NULL), exportWriter(NULL),
  system_loaded(false)
{
  cout << "> Creating Export1394 Page" << endl;

  entry_cip_n = GTK_ENTRY( lookup_widget( common->getWidget(), "entry_export_1394_cip_n") );
  entry_cip_d = GTK_ENTRY( lookup_widget( common->getWidget(), "entry_export_1394_cip_d") );
  entry_syt_offset = GTK_ENTRY( lookup_widget( common->getWidget(), "entry_export_1394_syt_offset") );
}

/** Destructor for page.
 */

Export1394::~Export1394() {
  cout << "> Destroying Export1394 Page" << endl;
}

/** Start of page.
 */

void Export1394::start() {
  static raw1394handle_t  handle;
  GtkLabel * load_label 
    = GTK_LABEL( lookup_widget ( exportPage->getWidget(),
				 "label_export_1394_load_messages" ) );
  
  gchar s[512];

  cout << ">> Starting Export1394" << endl;

  sprintf( s, "%d", Preferences::getInstance().cip_n );
  gtk_entry_set_text( entry_cip_n, s );
  sprintf( s, "%d", Preferences::getInstance().cip_d );
  gtk_entry_set_text( entry_cip_d, s );
  sprintf( s, "%d", Preferences::getInstance().syt_offset );
  gtk_entry_set_text( entry_syt_offset, s );

#ifdef LIBRAW1394_OLD
  if ((handle = raw1394_get_handle())) {
#else
  if ((handle = raw1394_new_handle())) {
#endif
    if (raw1394_set_port(handle, 0) >= 0) {
      raw1394_destroy_handle(handle);
      handle = NULL;
      exportReader = new IEEE1394Reader;
      exportReader->Open();
      system_loaded = true;
      gtk_label_set_text( load_label, "" );
    } else {
      system_loaded = false;
      gtk_label_set_text( load_label, 
			  "Error setting the IEEE 1394 port (host adapater); "
			  "all IEEE 1394 options are disabled.");
    }
  } else {
    system_loaded = false;
    gtk_label_set_text( load_label, 
			"The IEEE 1394 subsystem is not loaded; "
			"all IEEE 1394 options are disabled."); 
  }
  /* Change the sensitivity of the hosting widget according to 
     wheter the system loaded or not */
  GtkVBox * tmp 
    = GTK_VBOX( lookup_widget ( exportPage->getWidget(),
				"vbox_export_1394" ) );
  if (system_loaded) {
    gtk_widget_set_sensitive ( GTK_WIDGET ( tmp ), TRUE );
  } else {
    gtk_widget_set_sensitive ( GTK_WIDGET ( tmp ), FALSE );
  }
  
}
  
/** Define active widgets.
    This page does not support pausing, and if the system is
    not loaded, it does not support anything at all.
 */
  
gulong Export1394::onActivate() {
  if (system_loaded) {
    return EXPORT_PREVIEW | EXPORT_EXPORT | EXPORT_SCENE_LIST;
  } else {
    return 0;
  }
}
 
/** Leaving the page
 */
 
void Export1394::clean() {
  cout << ">> Leaving Export1394" << endl;
  if (exportReader) {
    exportReader->Close();
    delete exportReader;
    exportReader = NULL;
  }
  if (exportWriter) {
    delete exportWriter;
    exportWriter = NULL;
  }
  Preferences::getInstance().Save();
}
 
/** The actual export function */
 
void Export1394::doExport(PlayList * playlist, int begin, int end, int every,
			   bool preview) {
   Preferences::getInstance().cip_n = atoi( gtk_entry_get_text( entry_cip_n ));
   Preferences::getInstance().cip_d = atoi( gtk_entry_get_text( entry_cip_d ));
   Preferences::getInstance().syt_offset = atoi( gtk_entry_get_text( entry_syt_offset ));

   exportWriter = new IEEE1394Writer;
	
   // Only for export, NOT for preview. 
   if (!preview) {
     exportReader->AVCRecord( Preferences::getInstance().phyID );
   }
   
   static Frame frame;
   
   assert( exportWriter );
   if (exportWriter->viddev >= 0 ) {
     
     /* pre-roll */
     // TODO: Default frame? What is the intent of this?
     // common->getPlayList()->GetFrame( 0, frame );
     playlist->GetFrame( begin, frame );
     for (int i = begin; i <= end && i <= begin + 75 
	    && exportPage->isExporting; i++) {
       exportWriter->SendFrame( frame );
     }
     
     /* Iterate over all frames in selection */
     for (int i = begin; i <= end && exportPage->isExporting; i += every) {
       /* Call innerLoopUpdate */
       innerLoopUpdate(i, begin, end, every);
       
       /* Get the frame, write it */
       playlist->GetFrame( i, frame );
       exportWriter->SendFrame( frame );
     }
   }

   exportReader->AVCStop( Preferences::getInstance().phyID );
   
   delete exportWriter;
   exportWriter = NULL;
}
