/*
 * page_export_stills.cc Notebook Firewire/AVI/Still Frame Export Page Object
 * Copyright (C) 2001 Dan Dennedy <dan@dennedy.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <vector>
#include <iostream>
#include <strstream>
using std::cout;
using std::endl;
using std::strstream;

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <gnome.h>
#include <stdio.h>
#include <stdlib.h>

#include "kino_av_pipe.h"
#include "page_export_audio.h"
#include "preferences.h"
#include "kino_common.h"
#include "frame.h"
#include "page_editor.h"
#include "message.h"

extern "C" {
#include "support.h"
#include "callbacks.h"
#include "interface.h"
#include "commands.h"
}

/** Constructor for page.

\param common	common object to which this page belongs
*/

ExportAudio::ExportAudio( PageExport *_exportPage, KinoCommon *_common ) :
  Export(_exportPage, _common)
{
  cout << "> Creating ExportAudio Page" << endl;
  fileEntry 
    = GTK_ENTRY( gnome_file_entry_gtk_entry( GNOME_FILE_ENTRY( lookup_widget( common->getWidget(), 
									      "gnome_fileentry_export_audio" ) ) ) );
  
  sampleRate 
    = GTK_MENU( gtk_option_menu_get_menu( GTK_OPTION_MENU( lookup_widget( exportPage->getWidget(), 
									  "optionmenu_export_audio_sampling" ) ) ) );
  audioFormat
    = GTK_MENU( gtk_option_menu_get_menu( GTK_OPTION_MENU( lookup_widget( exportPage->getWidget(), 
									  "optionmenu_export_audio_format" ) ) ) );
  splitCheck
    = GTK_TOGGLE_BUTTON( lookup_widget( exportPage->getWidget(), 
					"checkbutton_export_audio_split" ) );	
}

/** Destructor for page.
 */

ExportAudio::~ExportAudio() {
  cout << "> Destroying ExportAudio Page" << endl;
}

/** start exporting audio
 */
void ExportAudio::doExport(PlayList * playlist, int begin, int end, int every,
			   bool preview) {
  cout << ">>> ExportAudio::doExport" << endl;
  static Frame *frame = new Frame;
  //TODO: Fix filename issues.
  char file[512];

  /* Check the filename */
  strcpy( file, gtk_entry_get_text( fileEntry ) );
  if ( !(strcmp(file, "") && strpbrk( file, " |`\"" ) == NULL) ) {
    modal_message( "You must enter a filename which doesn't contain special characters (spaces, quotes, pipes etc)." );
    return;
  }
  

  // Get a sample frame to obtain recording info
  playlist->GetFrame( begin, *frame );
  // TODO: Fix this, as this is testing....
  // ->GetFrame( atoi( gtk_entry_get_text( startSpin ) ) - 1, *frame );
  short channels = (short)frame->decoder->audio->num_channels;
  int inputFrequency = frame->decoder->audio->frequency;
  int outputFrequency = inputFrequency;
  bool splitScene = gtk_toggle_button_get_active( splitCheck );

  // Determine requested output frequency
  GtkWidget *active_item = gtk_menu_get_active( sampleRate );
  int menuValue = g_list_index (GTK_MENU_SHELL (sampleRate)->children, 
				active_item);
  if ( menuValue == 1 )
    outputFrequency = 32000;
  else if ( menuValue == 2 )
    outputFrequency = 44100;
  else if ( menuValue == 3 )
    outputFrequency = 48000;
  
  // Determine the output format
  active_item = gtk_menu_get_active( audioFormat );
  menuValue = g_list_index (GTK_MENU_SHELL (audioFormat)->children, active_item);
  
  char outputtemplate[ 1024 ];
  int resampleRate = 0;
  kino_audio_pipe type = PIPE_AUDIO_WAV;
  
  if ( menuValue == 0 ) {
    sprintf( outputtemplate, "%s%%s.wav", file );
    resampleRate = outputFrequency;
    type = PIPE_AUDIO_WAV;
  }
  else if ( menuValue == 1 ) {
    sprintf( outputtemplate, "| sox -t .raw -r %d -c 2 -w -s - -r %d %s%%s.wav", 
	     inputFrequency, outputFrequency, file );
    resampleRate = inputFrequency;
    type = PIPE_AUDIO_WAV;
  }
  else if ( menuValue == 2 ) {
    sprintf( outputtemplate, "| mp2enc -r %d -o %s%%s.mp2", 
	     outputFrequency, file );
    resampleRate = inputFrequency;
    type = PIPE_AUDIO_WAV;
  }
  else if ( menuValue == 3 ) {
    sprintf( outputtemplate, "| lame -r -s %d -x - -o %s%%s.mp3", 
	     inputFrequency / 1000, file );
    resampleRate = inputFrequency;
    type = PIPE_AUDIO_PCM;
  }
  else if ( menuValue == 4 ) {
    sprintf( outputtemplate, "| oggenc - -o %s%%s.ogg", file );
    resampleRate = inputFrequency;
    type = PIPE_AUDIO_WAV;
  }
  
  // Create the resampler and output objects
  AudioResample *resampler = new FastAudioResample( resampleRate );
  KinoAudioPipe *output = KinoAudioFactory::CreateAudioPipe( type );
  
  // Iterate through each scene (opening and closing audio pipes as requested)
  int scene = 0;
  
  for ( int sceneBegin = begin; sceneBegin <= end && exportPage->isExporting; 
	scene ++ ) {
    char full[ 1024 ];
    char sceneString[ 5 ];
    int sceneEnd = end;
    
    // Determine if we need to split by scene or not
    if ( splitScene ) {
      sprintf( sceneString, "%03d", scene );
      sceneEnd = playlist->FindEndOfScene( sceneBegin ) + 1;
      if ( sceneEnd > end )
	sceneEnd = end;
    }
    else {
      strcpy( sceneString, "" );
    }

    // expand the template to accomodate the scene splits
    sprintf( full, outputtemplate, sceneString );
    
    // output this scene
    bool active = output->OpenAudio( full, channels, resampleRate, 2 );
    if ( active ) {
      for (int i = sceneBegin; i <= sceneEnd && exportPage->isExporting; 
	   i += every ) {
	/* Call innerLoopUpdate */
	innerLoopUpdate(i, begin, end, every);
	
	playlist->GetFrame( i, *frame );
	resampler->Resample( *frame );
	if ( !output->OutputAudioFrame( resampler->output, resampler->size ) ) {
	  modal_message( "Error during audio export - aborting." );
	  exportPage->isExporting = false;
	}
	/* Update progressbar ... */
	exportPage->updateProgress( (gfloat) (i + every - begin)  
				    / (gfloat) (end + 1 - begin) );
      }
      output->CloseAudio();
    }
    else {
      modal_message( "Error starting output" );
      break;
    }
    
    // Move to start of next scene
    sceneBegin = sceneEnd + 1;
  }
  
  delete resampler;
  delete output;
}
