/*
 * page_export_avi.cc Notebook Firewire/AVI/Still Frame Export Page Object
 * Copyright (C) 2001 Dan Dennedy <dan@dennedy.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <iostream>
using std::cout;
using std::endl;

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <gnome.h>
#include <stdio.h>
#include <stdlib.h>

#include "page_export_avi.h"
#include "preferences.h"
#include "filehandler.h"
#include "message.h"
#include "page_editor.h"

extern "C" {
#include "support.h"
#include "callbacks.h"
#include "interface.h"
#include "commands.h"
}

/** Constructor for page.

\param _exportPage the export page object to bind to
\param _common	common object to which this page belongs
*/

ExportAVI::ExportAVI( PageExport *_exportPage, KinoCommon *_common ) : 
  Export(_exportPage, _common), writer(NULL)
{
  cout << "> Creating ExportAVI Page" << endl;

  /* Get pointers to our controls */
  fileEntry = GTK_ENTRY( gnome_file_entry_gtk_entry( GNOME_FILE_ENTRY( 
								      lookup_widget( common->getWidget(), 
										     "gnome_fileentry_export_avi" ) ) ) );
  formatDV1Toggle 
    = GTK_TOGGLE_BUTTON( lookup_widget( common->getWidget(), 
					"radiobutton_export_avi_type1") );
  formatDV2Toggle 
    = GTK_TOGGLE_BUTTON( lookup_widget( common->getWidget(), 
					"radiobutton_export_avi_type2") );
  formatRawDVToggle 
    = GTK_TOGGLE_BUTTON( lookup_widget( common->getWidget(), 
					"radiobutton_export_rawdv") );
  autoSplitToggle 
    = GTK_TOGGLE_BUTTON( lookup_widget( common->getWidget(), 
					"checkbutton_export_avi_autosplit" ) );
  timeStampToggle 
    = GTK_TOGGLE_BUTTON( lookup_widget( common->getWidget(),  
					"checkbutton_export_avi_timestamp" ) );
  framesEntry 
    = GTK_ENTRY( lookup_widget( exportPage->getWidget(), 
				"entry_export_avi_frames" ) );
  metadataToggle 
    = GTK_TOGGLE_BUTTON( lookup_widget( exportPage->getWidget(), 
				"checkbutton_export_avi_metadata" ) );	
  sizeEntry 
    = GTK_ENTRY( lookup_widget( exportPage->getWidget(), 
				"entry_export_avi_size" ) );	
  OpenDMLToggle 
    = GTK_TOGGLE_BUTTON( lookup_widget( common->getWidget(), 
					"checkbutton_export_avi_opendml") );

}

/** Destructor for page.
 */

ExportAVI::~ExportAVI() {
  cout << "> Destroying ExportAVI Page" << endl;
}

/** Start of page.
 */

void ExportAVI::start() {
  char s[20];
  cout << ">> Starting ExportAVI" << endl;
	
  switch (Preferences::getInstance().fileFormat ) {
  case AVI_DV1_FORMAT:
    gtk_toggle_button_set_active( formatDV1Toggle, TRUE );
    break;
  case AVI_DV2_FORMAT:
    gtk_toggle_button_set_active( formatDV2Toggle, TRUE);
    break;
  default:
    gtk_toggle_button_set_active( formatRawDVToggle, TRUE);
    break;
  }
  gtk_toggle_button_set_active( autoSplitToggle, 
				Preferences::getInstance().autoSplit);
  gtk_toggle_button_set_active( timeStampToggle, 
				Preferences::getInstance().timeStamp);
  gtk_entry_set_text( fileEntry, Preferences::getInstance().file);
  sprintf(s, "%d", Preferences::getInstance().frames);
  gtk_entry_set_text( framesEntry, s);
  sprintf(s, "%d", Preferences::getInstance().maxFileSize);
  gtk_entry_set_text( sizeEntry, s);
  gtk_toggle_button_set_active( OpenDMLToggle, Preferences::getInstance().isOpenDML);
}

/** Leaving the page
 */

void ExportAVI::clean() {
  cout << ">> Leaving ExportAVI" << endl;
  if (writer) {
    delete writer;
    writer = NULL;
  }
}


/** start exporting an AVI
 */
void ExportAVI::doExport(PlayList * playlist, int begin, int end, int every,
			 bool preview) {
  vector <int> scene = common->getPageEditor()->GetScene();
  int sceneIndex = 0;

  strcpy( Preferences::getInstance().file, gtk_entry_get_text( fileEntry ) );
	
  if ( !strcmp(Preferences::getInstance().file, "") ) { 
    modal_message( "You must enter a stem filename without an extension.");
    return;
  }

  bool updateMetadata = gtk_toggle_button_get_active( metadataToggle );
  
  if (gtk_toggle_button_get_active( formatDV1Toggle ))
    Preferences::getInstance().fileFormat = AVI_DV1_FORMAT;
  if (gtk_toggle_button_get_active( formatDV2Toggle ))
    Preferences::getInstance().fileFormat = AVI_DV2_FORMAT;
  if (gtk_toggle_button_get_active( formatRawDVToggle ))
    Preferences::getInstance().fileFormat = RAW_FORMAT;
  Preferences::getInstance().isOpenDML 
    = gtk_toggle_button_get_active( OpenDMLToggle );
  Preferences::getInstance().autoSplit 
    = gtk_toggle_button_get_active( autoSplitToggle );
  Preferences::getInstance().timeStamp 
    = gtk_toggle_button_get_active( timeStampToggle );
  Preferences::getInstance().frames = atoi( gtk_entry_get_text( framesEntry ) );
  Preferences::getInstance().maxFileSize = atoi( gtk_entry_get_text( sizeEntry ) );
  exportPage->resetProgress();
		
  switch (Preferences::getInstance().fileFormat) {
  case RAW_FORMAT:
    writer = new RawHandler();
    break;
				
  case AVI_DV1_FORMAT:
  case AVI_DV2_FORMAT:
    AVIHandler *aviWriter = new AVIHandler( Preferences::getInstance().fileFormat );
    writer = aviWriter;
    aviWriter->SetOpenDML( Preferences::getInstance().isOpenDML);
    break;
  }
  writer->SetMaxFrameCount( Preferences::getInstance().frames );
  writer->SetAutoSplit( false );
  writer->SetTimeStamp( Preferences::getInstance().timeStamp );
  // writer->SetEveryNthFrame( Preferences::getInstance().every );
  writer->SetEveryNthFrame( 1 );
  writer->SetMaxFileSize( (off_t)Preferences::getInstance().maxFileSize * (off_t)(1024*1024) );
  writer->SetBaseName( Preferences::getInstance().file );
	
  static Frame frame;
	
  try {
	  time_t datetime = time(NULL);
	  int frameNum = 0;
	  
    /* TODO: sample frame */
    playlist->GetFrame( begin, frame );
    writer->SetSampleFrame(frame);
    
    /* Iterate over all frames */
    for (int i = begin; i <= end && exportPage->isExporting; i += every) {
	
      /* Call innerLoopUpdate */
      innerLoopUpdate(i, begin, end, every);
	
      /* Extract frame, write stuff */
      playlist->GetFrame( i, frame );
      if ( Preferences::getInstance().autoSplit && writer->FileIsOpen()
	   && (i>0) && (i == scene[sceneIndex]) ) {
	sceneIndex++;
	writer->Close();
	frameNum = 0;
      }
	  if (updateMetadata)
	  {
		  frame.SetRecordingDate( &datetime, i );
		  frame.SetTimeCode( frameNum++ );
	  }
      writer->WriteFrame( frame );
    }
    writer->Close();
	
  } catch(string s) {
    cout << "Could not save AVI file " 
	 << Preferences::getInstance().file 
	 << ", because an exception has occurred: " << endl;
    cout << s << endl;
  }
  
  delete writer;
  writer = NULL;
}


